/*
********************************************************************************

This software module was originally developed by

Heiko Schwarz    (Fraunhofer HHI),
Tobias Hinz      (Fraunhofer HHI),
Karsten Suehring (Fraunhofer HHI)

in the course of development of the ISO/IEC 14496-10:2005 Amd.1 (Scalable Video
Coding) for reference purposes and its performance may not have been optimized.
This software module is an implementation of one or more tools as specified by
the ISO/IEC 14496-10:2005 Amd.1 (Scalable Video Coding).

Those intending to use this software module in products are advised that its
use may infringe existing patents. ISO/IEC have no liability for use of this
software module or modifications thereof.

Assurance that the originally developed software module can be used
(1) in the ISO/IEC 14496-10:2005 Amd.1 (Scalable Video Coding) once the
ISO/IEC 14496-10:2005 Amd.1 (Scalable Video Coding) has been adopted; and
(2) to develop the ISO/IEC 14496-10:2005 Amd.1 (Scalable Video Coding): 

To the extent that Fraunhofer HHI owns patent rights that would be required to
make, use, or sell the originally developed software module or portions thereof
included in the ISO/IEC 14496-10:2005 Amd.1 (Scalable Video Coding) in a
conforming product, Fraunhofer HHI will assure the ISO/IEC that it is willing
to negotiate licenses under reasonable and non-discriminatory terms and
conditions with applicants throughout the world.

Fraunhofer HHI retains full right to modify and use the code for its own
purpose, assign or donate the code to a third party and to inhibit third
parties from using the code for products that do not conform to MPEG-related
ITU Recommendations and/or ISO/IEC International Standards. 

This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 2005. 

********************************************************************************
*/


#include "H264AVCCommonLib.h"
#include "H264AVCCommonLib/PocCalculator.h"
#include <math.h>



H264AVC_NAMESPACE_BEGIN



PocCalculator::PocCalculator()
  : m_iLastIdrFieldNum    ( 0 )
  , m_iBitsLsb            ( 0 )
  , m_iTop2BotOffset      ( 0 )
  , m_iPrevRefPocMsb      ( 0 )
  , m_iPrevRefPocLsb      ( -1) //--ICU/ETRI FMO Implementation
  , m_iMaxPocLsb          ( 0 )
  , m_iFrameNumOffset     ( 0 )
  , m_iRefOffsetSum       ( 0 )
  , m_iPrevFrameNum       ( 0 )
{
}


ErrVal PocCalculator::create( PocCalculator*& rpcPocCalculator )
{
  rpcPocCalculator = new PocCalculator;

  ROT( NULL == rpcPocCalculator );

  return Err::m_nOK;
}


ErrVal PocCalculator::copy( PocCalculator*& rpcPocCalculator )
{
  rpcPocCalculator = new PocCalculator;

  ROT( NULL == rpcPocCalculator );

	rpcPocCalculator->m_iLastIdrFieldNum = m_iLastIdrFieldNum;
  rpcPocCalculator->m_iBitsLsb         = m_iBitsLsb;
  rpcPocCalculator->m_iTop2BotOffset   = m_iTop2BotOffset;
  rpcPocCalculator->m_iPrevRefPocMsb   = m_iPrevRefPocMsb;
  rpcPocCalculator->m_iPrevRefPocLsb   = m_iPrevRefPocLsb;
  rpcPocCalculator->m_iMaxPocLsb       = m_iMaxPocLsb;
  rpcPocCalculator->m_iFrameNumOffset  = m_iFrameNumOffset;
  rpcPocCalculator->m_iRefOffsetSum    = m_iRefOffsetSum;
  rpcPocCalculator->m_iPrevFrameNum    = m_iPrevFrameNum;

  return Err::m_nOK;
}


ErrVal PocCalculator::destroy()
{
  delete this;
  return Err::m_nOK;
}


ErrVal PocCalculator::calculatePoc( SliceHeader& rcSliceHeader )
{
  if( rcSliceHeader.getIdrFlag() )
  {
    RNOK( xInitSPS( rcSliceHeader.getSPS() ) );
  }

  switch( rcSliceHeader.getSPS().getPicOrderCntType() )
  {
  case 0:
    {
      //===== POC mode 0 =====
      Int iCurrPocMsb = m_iPrevRefPocMsb;
      Int iCurrPocLsb = rcSliceHeader.getPicOrderCntLsb();
      Int iDiffPocLsb = m_iPrevRefPocLsb - iCurrPocLsb;

      if( rcSliceHeader.getIdrFlag() )
      {
        iCurrPocMsb   = 0;
        iCurrPocLsb   = 0;
      }
      else if( iDiffPocLsb >= ( m_iMaxPocLsb >> 1 ) )
      {
        iCurrPocMsb  += m_iMaxPocLsb;
      }
      else if( -iDiffPocLsb > ( m_iMaxPocLsb >> 1 ) )
      {
        iCurrPocMsb  -= m_iMaxPocLsb;
      }
      if( rcSliceHeader.getNalRefIdc() )
      {
        m_iPrevRefPocMsb = iCurrPocMsb;
        m_iPrevRefPocLsb = iCurrPocLsb;
      }

			if( rcSliceHeader.getPicType() & TOP_FIELD )
			{
				rcSliceHeader.setTopFieldPoc( iCurrPocMsb + iCurrPocLsb );

				if( rcSliceHeader.getPicType() == FRAME )
				{
					rcSliceHeader.setBotFieldPoc( rcSliceHeader.getTopFieldPoc() + rcSliceHeader.getDeltaPicOrderCntBottom() );
				}
			}
			else
			{
				rcSliceHeader.setBotFieldPoc( iCurrPocMsb + iCurrPocLsb );
			}
    }
    break;
  case 1:
    {
      //===== POC mode 1 =====
      Int   iExpectedPoc    = 0;
      UInt  uiAbsFrameNum   = 0;

			//----- update parameters, set AbsFrameNum -----
			if( rcSliceHeader.getIdrFlag() )
			{
				m_iFrameNumOffset  = 0;
			}
			else if( m_iPrevFrameNum > (Int)rcSliceHeader.getFrameNum() )
      {
        m_iFrameNumOffset  += ( 1 << rcSliceHeader.getSPS().getLog2MaxFrameNum() );
	    }
      m_iPrevFrameNum       = rcSliceHeader.getFrameNum();
      if( rcSliceHeader.getSPS().getNumRefFramesInPicOrderCntCycle() )
      {
        uiAbsFrameNum       = m_iFrameNumOffset + m_iPrevFrameNum;
        if( uiAbsFrameNum > 0 && rcSliceHeader.getNalRefIdc() == 0 )
        {
          uiAbsFrameNum--;
        }
      }

      //--- get expected POC ---
      if( uiAbsFrameNum > 0 )
      {
        Int iPocCycleCount  = ( uiAbsFrameNum - 1 ) / rcSliceHeader.getSPS().getNumRefFramesInPicOrderCntCycle();
        Int iFrameNumCycle  = ( uiAbsFrameNum - 1 ) % rcSliceHeader.getSPS().getNumRefFramesInPicOrderCntCycle();
        iExpectedPoc        = iPocCycleCount * m_iRefOffsetSum;

        for( Int iIndex = 0; iIndex <= iFrameNumCycle; iIndex++ )
        {
          iExpectedPoc     += rcSliceHeader.getSPS().getOffsetForRefFrame( iIndex );
        }
      }
			else
			{
				iExpectedPoc = 0;
			}
      if( rcSliceHeader.getNalRefIdc() == 0 )
      {
        iExpectedPoc       += rcSliceHeader.getSPS().getOffsetForNonRefPic();
      }

			//----- set Poc -----
			if( rcSliceHeader.getPicType() & TOP_FIELD )
			{
				rcSliceHeader.setTopFieldPoc( iExpectedPoc + rcSliceHeader.getDeltaPicOrderCnt0() );

				if( rcSliceHeader.getPicType() == FRAME )
				{
					rcSliceHeader.setBotFieldPoc( rcSliceHeader.getTopFieldPoc() + rcSliceHeader.getDeltaPicOrderCnt1() + rcSliceHeader.getSPS().getOffsetForTopToBottomField() );
				}
			}
			else
			{
				rcSliceHeader.setBotFieldPoc( iExpectedPoc + rcSliceHeader.getDeltaPicOrderCnt0() + rcSliceHeader.getSPS().getOffsetForTopToBottomField() );
			}
    }
    break;
  case 2:
    {
      //===== POC mode 2 =====
      Int iCurrPoc;
			if( rcSliceHeader.getIdrFlag() )
			{
				m_iFrameNumOffset  = 0;
				iCurrPoc           = 0;
			}
			else
      {
        if( (Int)rcSliceHeader.getFrameNum() < m_iPrevFrameNum )
        {
          m_iFrameNumOffset += ( 1 << rcSliceHeader.getSPS().getLog2MaxFrameNum() );
        }
        if( rcSliceHeader.getNalRefIdc() )
        {
          iCurrPoc           = 2 * ( m_iFrameNumOffset + rcSliceHeader.getFrameNum() );
        }
        else
        {
          iCurrPoc           = 2 * ( m_iFrameNumOffset + rcSliceHeader.getFrameNum() ) - 1;
        }
      }
      if( rcSliceHeader.getNalRefIdc() )
      {
        m_iPrevFrameNum = rcSliceHeader.getFrameNum();
      }

			if( rcSliceHeader.getPicType() == FRAME )
			{
				rcSliceHeader.setTopFieldPoc( iCurrPoc );
				rcSliceHeader.setBotFieldPoc( iCurrPoc );
			}
			else if ( rcSliceHeader.getPicType() == TOP_FIELD )
			{
				rcSliceHeader.setTopFieldPoc( iCurrPoc );
			}
			else
			{
				rcSliceHeader.setBotFieldPoc( iCurrPoc );
			}
    }
    break;
  default:
    RERR();
    break;
  }
  return Err::m_nOK;
}

ErrVal
PocCalculator::resetMMCO5( SliceHeader& rcSliceHeader )
{
  PicType ePicType    = rcSliceHeader.getPicType();
  Int     iTempPOC    = ( ePicType == TOP_FIELD ? rcSliceHeader.getTopFieldPoc() :
                          ePicType == BOT_FIELD ? rcSliceHeader.getBotFieldPoc() :
                          min( rcSliceHeader.getTopFieldPoc(), rcSliceHeader.getBotFieldPoc() ) );
  Int     iTopFldPoc  = rcSliceHeader.getTopFieldPoc() - iTempPOC;
  Int     iBotFldPoc  = rcSliceHeader.getBotFieldPoc() - iTempPOC;
  m_iPrevRefPocMsb    = 0;
  m_iPrevRefPocLsb    = ( ePicType == BOT_FIELD ? iTopFldPoc : 0 );
  m_iFrameNumOffset   = 0;
  m_iPrevFrameNum     = 0;
  rcSliceHeader.setTopFieldPoc( iTopFldPoc );
  rcSliceHeader.setBotFieldPoc( iBotFldPoc );
  return Err::m_nOK;
}

ErrVal
PocCalculator::xInitSPS( const SequenceParameterSet& rcSPS )
{
  switch( rcSPS.getPicOrderCntType() )
  {
  case 0:
    {
      m_iPrevRefPocMsb  = 0;
      m_iPrevRefPocLsb  = 0;
      m_iMaxPocLsb      = ( 1 << rcSPS.getLog2MaxPicOrderCntLsb() );
    }
    break;
  case 1:
    {
      m_iFrameNumOffset = 0;
      m_iRefOffsetSum   = 0;
      for( UInt uiIndex = 0; uiIndex < rcSPS.getNumRefFramesInPicOrderCntCycle(); uiIndex++ )
      {
        m_iRefOffsetSum+= rcSPS.getOffsetForRefFrame( uiIndex );
      }
    }
    break;
  case 2:
    {
      m_iFrameNumOffset = 0;
    }
    break;
  default:
    RERR();
    break;
  }
  return Err::m_nOK;
}


ErrVal PocCalculator::setPoc( SliceHeader&  rcSliceHeader,
                              Int           iContNumber )
{
  ROTRS( iContNumber > ( INT_MAX - 1 ), Err::m_nERR );

  if( rcSliceHeader.getIdrFlag() )
  {
    m_iBitsLsb          = rcSliceHeader.getSPS().getLog2MaxPicOrderCntLsb();
    m_iLastIdrFieldNum  = iContNumber;
  }

  Int iCurrPoc = iContNumber - m_iLastIdrFieldNum;

	if( rcSliceHeader.getPicType() & TOP_FIELD )
	{
		rcSliceHeader.setTopFieldPoc  ( iCurrPoc );

		if( rcSliceHeader.getPicType() == FRAME )
		{
			rcSliceHeader.setBotFieldPoc( rcSliceHeader.getTopFieldPoc() + ( rcSliceHeader.getPPS().getPicOrderPresentFlag() ? m_iTop2BotOffset : 0 ) );
			rcSliceHeader.setDeltaPicOrderCntBottom ( rcSliceHeader.getBotFieldPoc() - rcSliceHeader.getTopFieldPoc() );
		}
	}
	else
	{
		rcSliceHeader.setBotFieldPoc  ( iCurrPoc );
	}
  rcSliceHeader.setPicOrderCntLsb ( iCurrPoc & ~( -1 << m_iBitsLsb ) );

  return Err::m_nOK;
}


H264AVC_NAMESPACE_END


